__name__ = "ammolite"
__author__ = "Patrick Kunzmann"
__all__ = ["get_and_set_pymol_instance",
           "launch_pymol", "launch_interactive_pymol", "reset",
           "setup_parameters", "DuplicatePyMOLException"]


_pymol = None


def get_and_set_pymol_instance(pymol_instance=None):
    """
    Get the global *PyMOL* instance.

    This function is intended for internal purposes and should only be 
    used for advanced usages.

    Parameters
    ----------
    pymol_instance : module or SingletonPyMOL, optional
        If a *PyMOL* instance is given here, the global instance is set
        to this instance.
        If *PyMOL* is already running and both instances are not the
        same, an exception is raised.
        By default *PyMOL* is started in library mode, if no *PyMOL*
        instance is currently running.
    
    Returns
    -------
    pymol_instance : module or SingletonPyMOL
        The global *PyMOL* instance.
    """
    global _pymol
    if pymol_instance is None:
        if not is_launched():
            _pymol = launch_pymol()
        return _pymol
    elif _pymol is None:
        if not hasattr(pymol_instance, "cmd"):
            raise TypeError("Given object is not a PyMOL instance")
        _pymol = pymol_instance
    elif _pymol is not pymol_instance:
        # Both the global pymol instance and the given instance are not
        # the same -> duplicate PyMOL instances
        raise DuplicatePyMOLException(
            "A PyMOL instance is already running"
        )
    return _pymol


def is_launched():
    """
    Check whether a *PyMOL* session is already running.

    Returns
    -------
    running : bool
        True, if a *PyMOL* instance is already running, false otherwise.
    """
    return _pymol is not None


def launch_pymol():
    """
    Launch *PyMOL* in object-oriented library mode.

    This is the recommended way to launch *PyMOL* if no GUI is
    required.
    This function simply creates a :class:`SingletonPyMOL` object,
    calls its :func:`start()` method and sets up necessary parameters using
    :func:`setup_parameters()`.

    Parameters
    ----------
    *args : str
        The command line options given to *PyMOL*.
    
    Returns
    -------
    pymol : SingletonPyMOL
        The started *PyMOL* instance.
        *PyMOL* commands can be invoked by using its :attr:`cmd`
        attribute.
    """
    from pymol2 import SingletonPyMOL
    global _pymol

    if is_launched():
        raise DuplicatePyMOLException(
            "A PyMOL instance is already running"
        )
    else:
        _pymol = SingletonPyMOL()
        _pymol.start()
        setup_parameters(_pymol)
    return _pymol


def launch_interactive_pymol(*args):
    """
    Launch a *PyMOL* GUI with the given command line arguments.

    It starts *PyMOL* by calling :func:`pymol.finish_launching()`,
    reinitializes *PyMOL* to clear the workspace and sets up necessary
    parameters using :func:`setup_parameters()`.

    Parameters
    ----------
    *args : str
        The command line options given to *PyMOL*.
    
    Returns
    -------
    pymol : module
        The :mod:`pymol` module.
        *PyMOL* commands can be invoked by using its :attr:`cmd`
        attribute.
    """
    import pymol
    global _pymol

    if is_launched():
        if _pymol is not pymol:
            raise DuplicatePyMOLException(
                "PyMOL is already running in library mode"
            )
        else:
            raise DuplicatePyMOLException(
                "A PyMOL instance is already running"
            )
    else:
        pymol.finish_launching(["pymol"] + list(args))
        _pymol = pymol
        pymol.cmd.reinitialize()
        setup_parameters(_pymol)
    return pymol


def reset():
    """
    Delete all objects in the PyMOL workspace and reset parameters to
    defaults.

    If *PyMOL* is not yet running, launch *PyMOL* in object-oriented
    library mode.
    """
    global _pymol

    if not is_launched(): 
        _pymol = launch_pymol()
    _pymol.cmd.reinitialize()
    setup_parameters(_pymol)


def setup_parameters(pymol_instance):
    """
    Sets *PyMOL* parameters that are necessary for *ammolite* to interact
    properly with *PyMOL*.

    pymol_instance : module or SingletonPyMOL or PyMOL, optional
        If *PyMOL* is used in library mode, the :class:`PyMOL`
        or :class:`SingletonPyMOL` object is given here.
        If otherwise *PyMOL* is used in GUI mode, the :mod:`pymol`
        module is given.
    """
    # The selections only work properly,
    # if the order stays the same after adding a model to PyMOL
    pymol_instance.cmd.set("retain_order", 1)


class DuplicatePyMOLException(Exception):
    pass


__name__ = "ammolite"
__author__ = "Patrick Kunzmann"
__all__ = ["draw_arrows"]

import numpy as np
from .cgo import draw_cgo, get_cylinder_cgo, get_cone_cgo


def draw_arrows(start, end, radius=0.1, head_radius=0.20,
                head_length=0.5, color=(0.5, 0.5, 0.5), head_color=None,
                name=None, pymol_instance=None):
    """
    Draw three-dimensional arrows using *Compiled Graphics Objects*
    (CGOs).

    Parameters
    ----------
    start, end : array-like, shape=(n,3)
        The start and end position of each arrow.
    radius, head_radius: float or array-like, shape=(n,), optional
        The radius of the tail and head for each arrow.
        Uniform for all arrows, if a single value is given.
    head_length: float or array-like, shape=(n,), optional
        The length of each arrow head.
        Uniform for all arrows, if a single value is given.
    color, head_color : array-like, shape=(3,) or shape=(n,3), optional
        The color of the tail and head for each arrow, given as RGB
        values in the range *(0, 1)*.
        Uniform for all arrows, if a single value is given.
        If no `head_color` is given, the arrows are single-colored.
    name : str, optional
        The name of the newly created CGO object.
        If omitted, a unique name is generated.
    pymol_instance : module or SingletonPyMOL or PyMOL, optional
        If *PyMOL* is used in library mode, the :class:`PyMOL`
        or :class:`SingletonPyMOL` object is given here.
        If otherwise *PyMOL* is used in GUI mode, the :mod:`pymol`
        module is given.
        By default the currently used *PyMOL* instance
        (``ammolite.pymol``) is used.
        If no *PyMOL* instance is currently running,
        *PyMOL* is started in library mode.
    """
    if head_color is None:
        head_color = color
    
    start = np.asarray(start)
    end = np.asarray(end)
    if start.ndim != 2 or end.ndim != 2:
        raise IndexError("Expected 2D array for start and end positions")
    if len(start) != len(end):
        raise IndexError(
            f"Got {len(start)} start positions, "
            f"but expected {len(end)} end positions"
        )
    expected_length = len(start)
    radius = _arrayfy(radius, expected_length, 1)
    head_radius = _arrayfy(head_radius, expected_length, 1)
    head_length = _arrayfy(head_length, expected_length, 1)
    color = _arrayfy(color, expected_length, 2)
    head_color = _arrayfy(head_color, expected_length, 2)
    
    normal = (end-start) / np.linalg.norm(end-start, axis=-1)[:, np.newaxis]
    middle = end - normal * head_length[:, np.newaxis]

    cgo_list = []
    for i in range(len(start)):
        cgo_list.extend([
            get_cylinder_cgo(
                start[i], middle[i], radius[i],
                color[i], color[i]
            ),
            get_cone_cgo(
                middle[i], end[i], head_radius[i], 0.0,
                head_color[i], head_color[i],
                True, False
            ),
        ])
    
    draw_cgo(cgo_list, name, pymol_instance)


def _arrayfy(value, length, min_dim):
    value = np.array(value, ndmin=min_dim)
    if len(value) == 1 and length > 1:
        value = np.repeat(value, length, axis=0)
    elif len(value) != length:
        raise IndexError(f"Expected {length} values, but got {len(value)}")
    return value


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Visualize structure data from Biotite with PyMOL.
"""

__version__ = "0.8.0"
__name__ = "ammolite"
__author__ = "Patrick Kunzmann"

from .cgo import *
from .convert import *
from .display import *
from .object import *
from .shapes import *
from .startup import *


# Make the PyMOL instance accessible via `ammolite.pymol`
# analogous to a '@property' of a class, but on module level instead
def __getattr__(name):
    if name == "pymol":
        return get_and_set_pymol_instance()
    elif name == "cmd":
        return __getattr__("pymol").cmd
    elif name in list(globals().keys()):
        return globals()["name"]
    else:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

def __dir__():
    return list(globals().keys()) + ["pymol", "cmd"]


__name__ = "ammolite"
__author__ = "Patrick Kunzmann"
__all__ = ["PyMOLObject", "NonexistentObjectError", "ModifiedObjectError"]

import numbers
from functools import wraps
import numpy as np
import biotite.structure as struc
from .convert import convert_to_atom_array, convert_to_chempy_model
from .startup import get_and_set_pymol_instance


def validate(method):
    """
    Check if the object name still exists and if the atom count has
    been modified.
    If this is the case, raise the appropriate exception.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self._check_existence()
        new_atom_count = self._cmd.count_atoms(f"model {self._name}")
        if new_atom_count != self._atom_count:
            raise ModifiedObjectError(
                f"The number of atoms in the object changed "
                f"from the original {self._atom_count} atoms "
                f" to {new_atom_count} atoms"
            )
        return method(self, *args, **kwargs)
    return wrapper


class PyMOLObject:
    """
    A wrapper around a *PyMOL object* (*PyMOL model*), usually created
    by the static :meth:`from_structure()` method.

    This class is primarily used to create *PyMOL* selection strings
    from boolean masks of an corresponding :class:`AtomArray` or
    :class:`AtomArrayStack` via the :meth:`where()`.
    Additionally, objects of this class provide wrapper methods for
    common *PyMOL* commands (e.g. ``show()`` or ``color()``), that
    directly support boolean masks for the ``selection`` parameter.

    Instances of this class become invalid, when atoms are added to or
    are deleted from the underlying *PyMOL* object.
    Calling methods of such an an invalidated object raises an
    :exc:`ModifiedObjectError`.
    Likewise, calling methods of an object, of which the underlying
    *PyMOL* object does not exist anymore, raises an
    :exc:`NonexistentObjectError`.

    Parameters
    ----------
    name : str
        The name of the *PyMOL* object.
    pymol_instance : module or SingletonPyMOL or PyMOL, optional
        If *PyMOL* is used in library mode, the :class:`PyMOL`
        or :class:`SingletonPyMOL` object is given here.
        If otherwise *PyMOL* is used in GUI mode, the :mod:`pymol`
        module is given.
        By default the currently used *PyMOL* instance
        (``ammolite.pymol``) is used.
        If no *PyMOL* instance is currently running,
        *PyMOL* is started in library mode.
    delete : PyMOL, optional
        If set to true, the underlying *PyMOL* object will be removed
        from the *PyMOL* session,
        when this object is garbage collected.
    
    Attributes
    ----------
    name : str
        The name of the *PyMOL* object.
    """
    
    _object_counter = 0
    _color_counter = 0
    

    def __init__(self, name, pymol_instance=None, delete=True):
        self._name = name
        self._pymol = get_and_set_pymol_instance(pymol_instance)
        self._delete = delete
        self._cmd = self._pymol.cmd
        self._check_existence()
        self._atom_count = self._cmd.count_atoms(f"model {self._name}")

    def __del__(self):
        if self._delete:
            try:
                # Try to delete this object from PyMOL
                # Fails if PyMOL itself is already garbage collected
                self._cmd.delete(self._name)
            except:
                pass



    @staticmethod
    def from_structure(atoms, name=None, pymol_instance=None, delete=True):
        """
        Create a :class:`PyMOLObject` from an :class:`AtomArray` or
        :class:`AtomArrayStack` and add it to the *PyMOL* session.

        Parameters
        ----------
        atoms : AtomArray or AtomArrayStack
            The structure to be converted.
        name : str, optional
            The name of the newly created *PyMOL* object.
            If omitted, a unique name is generated.
        pymol_instance : module or SingletonPyMOL or PyMOL, optional
            If *PyMOL* is used in library mode, the :class:`PyMOL`
            or :class:`SingletonPyMOL` object is given here.
            If otherwise *PyMOL* is used in GUI mode, the :mod:`pymol`
            module is given.
            By default the currently used *PyMOL* instance
            (``ammolite.pymol``) is used.
            If no *PyMOL* instance is currently running,
        *PyMOL* is started in library mode.
        delete : PyMOL, optional
            If set to true, the underlying *PyMOL* object will be
            removed from the *PyMOL* session, when this object is
            garbage collected.
        """
        pymol_instance = get_and_set_pymol_instance(pymol_instance)
        cmd = pymol_instance.cmd

        if name is None:
            name = f"ammolite_obj_{PyMOLObject._object_counter}"
            PyMOLObject._object_counter += 1
        
        if isinstance(atoms, struc.AtomArray) or \
        (isinstance(atoms, struc.AtomArrayStack) and atoms.stack_depth == 1):
                model = convert_to_chempy_model(atoms)
                cmd.load_model(model, name)
        elif isinstance(atoms, struc.AtomArrayStack):
            # Use first model as template
            model = convert_to_chempy_model(atoms[0])
            cmd.load_model(model, name)
            # Append states corresponding to all following models
            for coord in atoms.coord[1:]:
                cmd.load_coordset(coord, name)
        else:
            raise TypeError("Expected 'AtomArray' or 'AtomArrayStack'")

        return PyMOLObject(name, pymol_instance, delete)

    def to_structure(self, state=None, altloc="first", extra_fields=None,
                     include_bonds=False):
        """
        Convert this object into an :class:`AtomArray` or
        :class:`AtomArrayStack`.

        The returned :class:`AtomArray` contains the optional annotation
        categories ``b_factor``, ``occupancy`` and ``charge``.

        Parameters
        ----------
        state : int, optional
            If this parameter is given, the function will return an
            :class:`AtomArray` corresponding to the given state of the
            *PyMOL* object.
            If this parameter is omitted, an :class:`AtomArrayStack`
            containing all states will be returned, even if the *PyMOL*
            object contains only one state.
        altloc : {'first', 'occupancy', 'all'}
            This parameter defines how *altloc* IDs are handled:
            
            - ``'first'`` - Use atoms that have the first
              *altloc* ID appearing in a residue.
            - ``'occupancy'`` - Use atoms that have the *altloc* ID
              with the highest occupancy for a residue.
            - ``'all'`` - Use all atoms.
              Note that this leads to duplicate atoms.
              When this option is chosen, the ``altloc_id``
              annotation array is added to the returned structure.
            
        include_bonds : bool, optional
            If set to true, an associated :class:`BondList` will be created
            for the returned structure.
        
        Returns
        -------
        structure : AtomArray or AtomArrayStack
            The converted structure.
            Whether an :class:`AtomArray` or :class:`AtomArrayStack` is
            returned depends on the `state` parameter.
        """
        if state is None:
            model = self._cmd.get_model(self._name, state=1)
            template = convert_to_atom_array(
                model, include_bonds
            )
            expected_length = None
            coord = []
            for i in range(self._cmd.count_states(self._name)):
                state_coord = self._cmd.get_coordset(self._name, state=i+1)
                if expected_length is None:
                    expected_length = len(state_coord)
                elif len(state_coord) != expected_length:
                    raise ValueError(
                        "The models have different numbers of atoms"
                    )
                coord.append(state_coord)
            coord = np.stack(coord)
            structure = struc.from_template(template, coord)
        
        else:
            model = self._cmd.get_model(self._name, state=state)
            structure = convert_to_atom_array(
                model, include_bonds
            )
        
        # Filter altloc IDs and return
        if altloc == "occupancy":
            structure = structure[
                ...,
                struc.filter_highest_occupancy_altloc(
                    structure, structure.altloc_id, structure.occupancy
                )
            ]
            structure.del_annotation("altloc_id")
            return structure
        elif altloc == "first":
            structure = structure[
                ...,
                struc.filter_first_altloc(structure, structure.altloc_id)
            ]
            structure.del_annotation("altloc_id")
            return structure
        elif altloc == "all":
            return structure
        else:
            raise ValueError(f"'{altloc}' is not a valid 'altloc' option")


    
    @property
    def name(self):
        return self._name
    

    def exists(self):
        """
        Check whether the underlying *PyMOL* object still exists.

        Returns
        -------
        bool
            True if the *PyMOL* session contains an object with the name
            of this :class:`PyMOLObject`, false otherwise.
        """
        return self._name in self._cmd.get_object_list()

    def _check_existence(self):
        if not self.exists():
            raise NonexistentObjectError(
                f"A PyMOL object with the name {self._name} "
                f"does not exist anymore"
            )


    @validate
    def where(self, index):
        """
        Convert a *Biotite*-compatible atom selection index
        (integer, slice, boolean mask, index array) into a *PyMOL*
        selection expression.

        Parameters
        ----------
        index : int or slice or ndarray, dtype=bool or ndarray, dtype=int
            The boolean mask to be converted into a selection string.
        
        Returns
        -------
        expression : str
            A *PyMOL* compatible selection expression.
        """
        if isinstance(index, numbers.Integral):
            # PyMOLs indexing starts at 1
            return f"model {self._name} and index {index}"
        
        elif isinstance(index, np.ndarray) and index.dtype == bool:
            mask = index
            if len(mask) != self._atom_count:
                raise IndexError(
                    f"Mask has length {len(mask)}, but the number of "
                    f"atoms in the PyMOL model is {self._atom_count}"
                )
        
        else:
            # Convert any other index type into a boolean mask
            mask = np.zeros(self._atom_count, dtype=bool)
            mask[index] = True
        
        # Indices where the mask changes from True to False
        # or from False to True
        # The '+1' makes each index refer to the position
        # after the change i.e. the new value
        changes = np.where(np.diff(mask))[0] + 1
        # If first element is True, insert index 0 at start
        # -> the first change is always from False to True
        if mask[0]:
            changes = np.concatenate(([0], changes))
        # If the last element is True, insert append length of mask
        # as exclusive stop index
        # -> the last change is always from True to False
        if mask[-1]:
            changes = np.concatenate((changes, [len(mask)]))
        # -> Changes are alternating (F->T, T->F, F->T, ..., F->T, T->F)
        # Reshape into pairs ([F->T, T->F], [F->T, T->F], ...)
        # -> these are the intervals where the mask is True
        intervals = np.reshape(changes, (-1, 2))

        if len(intervals) > 0:
            # Convert interval into selection string
            # Two things to note:
            # - PyMOLs indexing starts at 1-> 'start+1'
            # - Stop in 'intervals' is exclusive -> 'stop+1-1' -> 'stop'
            index_selection = " or ".join(
                [f"index {start+1}-{stop}" for start, stop in intervals]
            )
            # Constrain the selection to given object name
            return f"model {self._name} and ({index_selection})"
        else:
            return "none"
    
    def _into_selection(self, selection, not_none=False):
        """
        Turn a boolean mask into a *PyMOL* selection expression or 
        restrict an selection expression to the current *PyMOL* object.
        """
        if selection is None:
            return f"model {self._name}"
        elif isinstance(selection, str):
            return f"model {self._name} and ({selection})"
        else:
            sel = self.where(np.asarray(selection))
            if sel == "none" and not_none:
                raise ValueError("Selection contains no atoms")
            return sel




    @validate
    def alter(self, selection, expression):
        """
        Change atomic properties using an expression evaluated
        within a temporary namespace for each atom.

        This method is a thin wrapper around the *PyMOL* ``alter()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
        expression : str
            The properties of the selected atoms are changed based on
            this expression.
        """
        self._cmd.alter(self._into_selection(selection), expression)
    
    @validate
    def cartoon(self, type, selection=None):
        """
        Change the default cartoon representation for a selection
        of atoms.

        This method is a thin wrapper around the *PyMOL* ``cartoon()``
        command.

        Parameters
        ----------
        type : str
            One of
            
            - ``'automatic'``,
            - ``'skip'``,
            - ``'loop'``,
            - ``'rectangle'``,
            - ``'oval'``,
            - ``'tube'``,
            - ``'arrow'`` or
            - ``'dumbbell'``.
        
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.cartoon(type, self._into_selection(selection))
    
    @validate
    def center(self, selection=None, state=None, origin=True):
        """
        Translate the window, the clipping slab, and the
        origin to a point centered within the atom selection.

        This method is a thin wrapper around the *PyMOL* ``center()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        origin : bool, optional
            If set to false, the origin is left unchanged.
        """
        state = 0 if state is None else state
        self._cmd.center(self._into_selection(selection), state, int(origin))

    @validate
    def clip(self, mode, distance, selection=None, state=None):
        """
        Alter the positions of the near and far clipping planes.

        This method is a thin wrapper around the *PyMOL* ``clip()``
        command.

        Parameters
        ----------
        mode : {'near', 'far', 'move', 'slab', 'atoms'}
            
            - ``near`` - Move the near plane
            - ``far`` - Move the far plane
            - ``move`` - Move slab
            - ``slab`` - Set slab thickness
            - ``atoms`` - clip selected atoms with the given buffer

        distance : float
            The meaning of this parameter depends on `mode`.
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.clip(mode, distance, self._into_selection(selection), state)
    
    @validate
    def color(self, color, selection=None, representation=None):
        """
        Change the color of atoms.

        This method is a thin wrapper around the *PyMOL* ``color()``
        or ``set("xxx_color")`` command.

        Parameters
        ----------
        color : str or tuple(float, float, float)
            Either a *PyMOL* color name or a tuple containing an RGB
            value (0.0 to 1.0).
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        representation : {"sphere", "surface", "mesh", "dot", "cartoon", "ribbon"}, optional
            Colors only the given representation by internally calling
            ``set("xxx_color")``.
            By default, all representations are affected, i.e. ``color()``
            is called internally.
        
        Notes
        -----
        If an RGB color is given, the color is registered as a unique
        named color via the ``set_color()`` command.
        """
        if not isinstance(color, str):
            color_name = f"ammolite_color_{PyMOLObject._color_counter}"
            PyMOLObject._color_counter += 1
            self._cmd.set_color(color_name, tuple(color))
        else:
            color_name = color
            registered = [name for name, _ in self._cmd.get_color_indices()]
            if color_name not in registered:
                raise ValueError(
                    f"Unknown color '{color_name}'"
                )
        
        if representation is None:
            self._cmd.color(color_name, self._into_selection(selection))
        else:
            if representation not in (
                "sphere", "surface", "mesh", "dot", "cartoon", "ribbon"
            ):
                raise ValueError(
                    f"'{representation}' is not a supported representation"
                )
            self._cmd.set(
                f"{representation}_color",
                color_name, self._into_selection(selection)
            )
    
    
    @validate
    def surface_color(self, color, selection=None):
        """
        Change the color of displayed surfaces.

        This method is a thin wrapper around *PyMOL*
        ``set("surface_color")``.

        Parameters
        ----------
        color : str or tuple(float, float, float)
            Either a *PyMOL* color name or a tuple containing an RGB
            value (0.0 to 1.0).
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        
        Notes
        -----
        If an RGB color is given, the color is registered as a unique
        named color via the ``set_color()`` command.
        """
        color_name = self._created_named_color(color)
        self._cmd.set(
            "surface_color", color_name, self._into_selection(selection)
        )

    @validate
    def desaturate(self, selection=None, a=0.5):
        """
        Desaturate the colors of the selected atoms.

        This method is a thin wrapper around the *PyMOL*
        ``desaturate()`` command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        a : float
            A desaturation factor between 0.0 and 1.0.
        """
        self._cmd.desaturate(self._into_selection(selection), a)

    @validate
    def disable(self, selection=None):
        """
        Turn off display of the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``disable()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.disable(self._into_selection(selection))
    
    @validate
    def distance(self, name, selection1, selection2, cutoff=None, mode=None,
                 show_label=True, width=None, length=None, gap=None):
        """
        Create a new distance object between two atom selections.

        This method is a thin wrapper around the *PyMOL* ``distance()``
        command.

        Parameters
        ----------
        name : str
            Name of the distance object to create.
        selection1, selection2 : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
        cutoff : float, optional
            The longest distance to show.
        mode: {0, 1, 2, 3, 4}, optional
            
            - ``0`` - All interatomic distances
            - ``1`` - Only bond distances
            - ``2`` - Only polar contact distances
            - ``3`` - All interatomic distances,
              use distance_exclusion setting
            - ``4`` - Distance between centroids

        show_label : bool, optional
            Whether to show the distance as label.
        width, length, gap : float optional
            The width and length of each dash and the gap length between
            the dashes.
        """
        self._cmd.distance(
            name,
            self._into_selection(selection1),
            self._into_selection(selection2),
            cutoff,
            mode,
            label=int(show_label),
            width=width,
            length=length,
            gap=gap
        )
    
    @validate
    def dss(self, selection=None, state=None):
        """
        Determine the secondary structure of the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``dss()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.dss(self._into_selection(selection), state)
    
    @validate
    def enable(self, selection=None):
        """
        Turn on display of the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``enable()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.enable(self._into_selection(selection))

    @validate
    def hide(self, representation, selection=None):
        """
        Turn off an atom representation (e.g. sticks, spheres, etc.).

        This method is a thin wrapper around the *PyMOL* ``hide()``
        command.

        Parameters
        ----------
        representation : str
            One of
            
            - ``'lines'``,
            - ``'spheres'``,
            - ``'mesh'``,
            - ``'ribbon'``,
            - ``'cartoon'``,
            - ``'sticks'``,
            - ``'dots'``,
            - ``'surface'``,
            - ``'label'``,
            - ``'extent'``,
            - ``'nonbonded'``,
            - ``'nb_spheres'``,
            - ``'slice'`` or
            - ``'cell'``.

        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.hide(representation, self._into_selection(selection))
    
    @validate
    def indicate(self, selection=None):
        """
        Show a visual representation of the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``indicate()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.indicate(self._into_selection(selection))
    
    @validate
    def label(self, selection, text):
        """
        Label the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``label()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
        text : str
            The label text.
        """
        self._cmd.label(self._into_selection(selection), f'"{text}"')
    
    @validate
    def orient(self, selection=None, state=None):
        """
        Align the principal components of the selected atoms with the
        *xyz* axes.

        This method is a thin wrapper around the *PyMOL* ``orient()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.orient(self._into_selection(selection, True), state)
    
    @validate
    def origin(self, selection=None, state=None):
        """
        Set the center of rotation about the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``origin()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.origin(
            selection=self._into_selection(selection), state=state
        )
    
    @validate
    def select(self, name, selection=None):
        """
        Create a named selection object from the selected atoms.

        This method is a thin wrapper around the *PyMOL* ``select()``
        command.

        Parameters
        ----------
        name : str
            Name of the selection object to create.
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.select(name, self._into_selection(selection))

    @validate
    def set(self, name, value, selection=None, state=None):
        """
        Change per-atom settings.

        This method is a thin wrapper around the *PyMOL* ``set()``
        command.

        Parameters
        ----------
        name : str
            The name of the setting to be changed.
            One of

            - ``'sphere_color'``,
            - ``'surface_color'``,
            - ``'mesh_color'``,
            - ``'label_color'``,
            - ``'dot_color'``,
            - ``'cartoon_color'``,
            - ``'ribbon_color'``,
            - ``'transparency'`` (for surfaces) or
            - ``'sphere_transparency'``.

        value : object
            The new value for the given setting name.
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.set(name, value, self._into_selection(selection), state)

    @validate
    def set_bond(self, name, value, selection1=None, selection2=None,
                 state=None):
        """
        Change per-bond settings for all bonds which exist
        between two atom selections.

        This method is a thin wrapper around the *PyMOL* ``set_bond()``
        command.

        Parameters
        ----------
        name : str
            The name of the setting to be changed.
            One of

            - ``'valence'``,
            - ``'line_width'``,
            - ``'line_color'``,
            - ``'stick_radius'``,
            - ``'stick_color'`` or
            - ``'stick_transparency'``.

        value : object
            The new value for the given setting name.
        selection1, selection2 : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, `selection1` applies to all atoms of this
            *PyMOL* object and `selection2` applies to the same atoms as
            `selection1`.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        selection2 = selection1 if selection2 is None else selection2
        self._cmd.set_bond(
            name, value,
            self._into_selection(selection1), self._into_selection(selection2),
            state
        )
    
    @validate
    def show(self, representation, selection=None):
        """
        Turn on an atom representation (e.g. sticks, spheres, etc.).

        This method is a thin wrapper around the *PyMOL* ``show()``
        command.

        Parameters
        ----------
        representation : str
            One of

            - ``'lines'``,
            - ``'spheres'``,
            - ``'mesh'``,
            - ``'ribbon'``,
            - ``'cartoon'``,
            - ``'sticks'``,
            - ``'dots'``,
            - ``'surface'``,
            - ``'label'``,
            - ``'extent'``,
            - ``'nonbonded'``,
            - ``'nb_spheres'``,
            - ``'slice'`` or
            - ``'cell'``.

        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.show(representation, self._into_selection(selection))

    @validate
    def show_as(self, representation, selection=None):
        """
        Turn on a representation (e.g. sticks, spheres, etc.) and hide
        all other representations.

        This method is a thin wrapper around the *PyMOL* ``show_as()``
        command.

        Parameters
        ----------
        representation : str
            One of

            - ``'lines'``,
            - ``'spheres'``,
            - ``'mesh'``,
            - ``'ribbon'``,
            - ``'cartoon'``,
            - ``'sticks'``,
            - ``'dots'``,
            - ``'surface'``,
            - ``'label'``,
            - ``'extent'``,
            - ``'nonbonded'``,
            - ``'nb_spheres'``,
            - ``'slice'`` or
            - ``'cell'``.

        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        """
        self._cmd.show_as(representation, self._into_selection(selection))
    
    @validate
    def smooth(self, selection=None, passes=1, window=5,
               first=1, last=0, ends=False):
        """
        Perform a moving average over the coordinate states.

        This method is a thin wrapper around the *PyMOL* ``smooth()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        passes : int, optional
            The number of smoothing passes.
        windows : int, optional
            The size of the moving window.
        first, last : int, optional
            The interval of states to smooth.
        ends : bool, optional
            If set to true, the end states are also smoothed using a
            weighted asymmetric window.
        """
        self._cmd.smooth(
            self._into_selection(selection), passes, window,
            first, last, int(ends)
        )
    
    # TODO: def spectrum()
    
    @validate
    def unset(self, name, selection=None, state=None):
        """
        Clear per-atom settings.

        This method is a thin wrapper around the *PyMOL* ``set()``
        command.

        Parameters
        ----------
        name : str
            The name of the setting to be cleared.
            One of

            - ``'sphere_color'``,
            - ``'surface_color'``,
            - ``'mesh_color'``,
            - ``'label_color'``,
            - ``'dot_color'``,
            - ``'cartoon_color'``,
            - ``'ribbon_color'``,
            - ``'transparency'`` (for surfaces) or
            - ``'sphere_transparency'``.

        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        self._cmd.unset(name, self._into_selection(selection), state)

    @validate
    def unset_bond(self, name, selection1=None, selection2=None, state=None):
        """
        Clear per-bond settings for all bonds which exist
        between two atom selections.

        This method is a thin wrapper around the *PyMOL* ``unset_bond()``
        command.

        Parameters
        ----------
        name : str
            The name of the setting to be cleared.
            One of

            - ``'valence'``,
            - ``'line_width'``,
            - ``'line_color'``,
            - ``'stick_radius'``,
            - ``'stick_color'`` or
            - ``'stick_transparency'``.

        selection1, selection2 : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, `selection1` applies to all atoms of this
            *PyMOL* object and `selection2` applies to the same atoms as
            `selection1`.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        """
        state = 0 if state is None else state
        selection2 = selection1 if selection2 is None else selection2
        self._cmd.unset_bond(
            name,
            self._into_selection(selection1), self._into_selection(selection2),
            state
        )
    
    @validate
    def zoom(self, selection=None, buffer=0.0, state=None, complete=False):
        """
        Scale and translate the window and the origin to cover the
        selected atoms.

        This method is a thin wrapper around the *PyMOL* ``zoom()``
        command.

        Parameters
        ----------
        selection : str or int or slice or ndarray, dtype=bool or ndarray, dtype=int, optional
            A *Biotite* compatible atom selection index,
            e.g. a boolean mask, or a *PyMOL* selection expression that
            selects the atoms of this *PyMOL* object to apply the
            command on.
            By default, the command is applied on all atoms of this
            *PyMOL* object.
        buffer : float, optional
            An additional distance to the calculated camera position.
        state : int, optional
            The state to apply the command on.
            By default, the command is applied on all states of this
            *PyMOL* object.
        complete : bool, optional
            If set to true, it is insured that no atoms centers are
            clipped.
        """
        state = 0 if state is None else state
        self._cmd.zoom(
            self._into_selection(selection, True),
            buffer,
            state,
            int(complete)
        )


class NonexistentObjectError(Exception):
    """
    Indicates that a *PyMOL* object with a given name does not exist.
    """
    pass

class ModifiedObjectError(Exception):
    """
    Indicates that a atoms were added to or removed from the *PyMOL*
    object after the corresponding :class:`PyMOLObject` was created.
    """
    pass


__name__ = "ammolite"
__author__ = "Patrick Kunzmann"
__all__ = ["convert_to_atom_array", "convert_to_chempy_model"]

import warnings
import numpy as np
from biotite.sequence import ProteinSequence
import biotite.structure as struc
from chempy.models import Indexed as IndexedModel
from chempy import Atom, Bond


BOND_ORDER = {
    struc.BondType.ANY: 1,
    struc.BondType.SINGLE: 1,
    struc.BondType.DOUBLE: 2,
    struc.BondType.TRIPLE: 3,
    struc.BondType.QUADRUPLE: 4,
    struc.BondType.AROMATIC_SINGLE: 1,
    struc.BondType.AROMATIC_DOUBLE: 2,
}


def convert_to_atom_array(chempy_model, include_bonds=False):
    """
    Convert a :class:`chempy.models.Indexed`
    object into an :class:`AtomArray`.

    The returned :class:`AtomArray` contains the optional annotation
    categories ``b_factor``, ``occupancy``, ``charge`` and
    ``altloc_id``.
    No *altloc* ID filtering is performed.

    Parameters
    ----------
    chempy_model : Indexed
        The ``chempy`` model.
    include_bonds : bool, optional
        If set to true, an associated :class:`BondList` will be created
        for the returned atom array.
    
    Returns
    -------
    atom_array : AtomArray
        The converted structure.
    """
    atoms = chempy_model.atom
    
    bonds = chempy_model.bond

    
    atom_array = struc.AtomArray(len(atoms))
    
    
    # Add annotation arrays
    atom_array.chain_id = np.array(
        [a.chain for a in atoms],
        dtype="U3"
    )
    atom_array.res_id = np.array(
        [a.resi_number for a in atoms],
        dtype=int
    )
    atom_array.ins_code = np.array(
        [a.ins_code for a in atoms],
        dtype="U1"
    )
    atom_array.res_name = np.array(
        [a.resn for a in atoms],
        dtype="U3"
    )
    atom_array.hetero = np.array(
        [a.hetatm for a in atoms],
        dtype=bool
    )
    atom_array.atom_name = np.array(
        [a.name for a in atoms],
        dtype="U6"
    )
    atom_array.element = np.array(
        [a.symbol for a in atoms],
        dtype="U2"
    )
    
    atom_array.set_annotation(
        "b_factor",
        np.array(
            [a.b if hasattr(a, "b") else 0
             for a in atoms],
            dtype=float
        )
    )
    atom_array.set_annotation(
        "occupancy",
        np.array(
            [a.q if hasattr(a, "q") else 1.0 for a in atoms],
            dtype=float
        )
    )
    atom_array.set_annotation(
        "charge",
        np.array(
            [a.formal_charge if hasattr(a, "formal_charge") else 0
             for a in atoms],
            dtype=int
        )
    )
    atom_array.set_annotation(
        "altloc_id",
        np.array(
            [a.alt if hasattr(a, "alt") else "" for a in atoms],
            dtype="U1"
        )
    )
    

    # Set coordinates
    atom_array.coord = np.array(
        [a.coord for a in atoms],
        dtype=np.float32
    )


    # Add bonds
    if include_bonds:
        bond_array = np.array(
            [[b.index[0], b.index[1], b.order] for b in bonds],
            dtype=np.uint32
        )
        atom_array.bonds = struc.BondList(len(atoms), bond_array)
    

    return atom_array


def convert_to_chempy_model(atom_array):
    """
    Convert an :class:`AtomArray` into a :class:`chempy.models.Indexed`
    object.

    Returns
    -------
    chempy_model : Indexed
        The converted structure.
    """
    model = IndexedModel()

    annot_cat = atom_array.get_annotation_categories()


    for i in range(atom_array.array_length()):
        atom = Atom()

        atom.segi = atom_array.chain_id[i]
        atom.chain = atom_array.chain_id[i]

        atom.resi_number = atom_array.res_id[i]

        atom.ins_code = atom_array.ins_code[i]
        
        res_name = atom_array.res_name[i]
        atom.resn = res_name
        if len(res_name) == 1:
            atom.resn_code = res_name
        else:
            try:
                atom.resn_code = ProteinSequence.convert_letter_3to1(res_name)
            except KeyError:
                atom.resn_code = "X"

        atom.hetatm = 1 if atom_array.hetero[i] else 0

        atom.name = atom_array.atom_name[i]

        atom.symbol = atom_array.element[i]
        
        if "b_factor" in annot_cat:
            atom.b = atom_array.b_factor[i]
        
        if "occupancy" in annot_cat:
            atom.q = atom_array.occupancy[i]
        
        if "charge" in annot_cat:
            atom.formal_charge = atom_array.charge[i]
        
        atom.coord = tuple(atom_array.coord[..., i, :])

        atom.index = i+1

        model.add_atom(atom)
    

    if atom_array.bonds is not None:
        for i, j, bond_type in atom_array.bonds.as_array():
            bond = Bond()
            bond.order = BOND_ORDER[bond_type]
            bond.index = [i, j]
            model.add_bond(bond)
    else:
        warnings.warn(
            "The given atom array (stack) has no associated bond information"
        )
    

    return model


__name__ = "ammolite"
__author__ = "Patrick Kunzmann"
__all__ = ["draw_cgo", "get_cylinder_cgo", "get_cone_cgo", "get_sphere_cgo"]

import itertools
import numpy as np
from .startup import get_and_set_pymol_instance


_object_counter = 0


def draw_cgo(cgo_list, name=None, pymol_instance=None):
    """
    Draw geometric shapes using *Compiled Graphics Objects* (CGOs).

    Each CGO is represented by a list of floats, which can be obtained
    via the ``get_xxx_cgo()`` functions.

    Parameters
    ----------
    cgo_list : list of list of float
        The CGOs to draw.
        It is recommended to use a ``get_xxx_cgo()`` function to obtain
        the elements for this list, if possible.
        Otherwise, shapes may be drawn incorrectly or omitted entirely,
        if a CGO is incorrectly formatted.
    name : str, optional
        The name of the newly created CGO object.
        If omitted, a unique name is generated.
    pymol_instance : module or SingletonPyMOL or PyMOL, optional
        If *PyMOL* is used in library mode, the :class:`PyMOL`
        or :class:`SingletonPyMOL` object is given here.
        If otherwise *PyMOL* is used in GUI mode, the :mod:`pymol`
        module is given.
        By default the currently used *PyMOL* instance
        (``ammolite.pymol``) is used.
        If no *PyMOL* instance is currently running,
        *PyMOL* is started in library mode.
    """
    global _object_counter
    if name is None:
        name = f"ammolite_cgo_{_object_counter}"
        _object_counter += 1
    pymol_instance = get_and_set_pymol_instance(pymol_instance)
    pymol_instance.cmd.load_cgo(list(itertools.chain(*cgo_list)), name)


def get_cylinder_cgo(start, end, radius, start_color, end_color):
    """
    Get the CGO for a cylinder.

    Parameters
    ----------
    start, end : array-like, shape=(3,)
        The start and end position of the cylinder.
    radius : float
        The radius of the cylinder
    start_color, end_color : array-like, shape=(3,)
        The color at the start and end of the cylinder given as RGB
        values in the range *(0, 1)*.
    """
    CYLINDER = 9.0
    _expect_length(start, "start", 3)
    _expect_length(end, "end", 3)
    _expect_length(start_color, "start_color", 3)
    _expect_length(end_color, "end_color", 3)
    _check_color(start_color)
    _check_color(end_color)
    return [
        CYLINDER, *start, *end, radius, *start_color, *end_color
    ]

def get_cone_cgo(start, end, start_radius, end_radius,
                  start_color, end_color, start_cap, end_cap):
    """
    Get the CGO for a cone.

    Parameters
    ----------
    start, end : array-like, shape=(3,)
        The start and end position of the cone.
    start_radius, end_radius : float
        The radius of the cone at the start and end.
    start_color, end_color : array-like, shape=(3,)
        The color at the start and end of the cone given as RGB
        values in the range *(0, 1)*.
    start_cap, end_cap : bool
        If true, a cap is drawn at the start or end of the cone.
        Otherwise the cone is displayed as *open*.
    """
    CONE = 27.0
    _expect_length(start, "start", 3)
    _expect_length(end, "end", 3)
    _expect_length(start_color, "start_color", 3)
    _expect_length(end_color, "end_color", 3)
    _check_color(start_color)
    _check_color(end_color)
    return [
        CONE, *start, *end, start_radius, end_radius, *start_color, *end_color,
        float(start_cap), float(end_cap)
    ]

def get_sphere_cgo(pos, radius, color):
    """
    Get the CGO for a sphere.

    Parameters
    ----------
    pos : array-like, shape=(3,)
        The position of the sphere.
    radius : float
        The radius of the sphere.
    color : array-like, shape=(3,)
        The color of the sphere given as RGB values in the range
        *(0, 1)*.
    """
    COLOR = 6.0
    SPHERE = 7.0
    _expect_length(pos, "pos", 3)
    _expect_length(color, "color", 3)
    _check_color(color)
    return [
        COLOR, *color, SPHERE, *pos, radius
    ]


def _expect_length(values, name, length):
    if len(values) != length:
        raise IndexError(
            f"'{name}' has {len(values)} values, but {length} were expected"
        )


def _check_color(color):
    if np.any(color) < 0 or np.any(color) > 1:
        raise ValueError("Colors must be in range (0, 1)")